==> wrogue-0.7.4.tar.bz2: wrogue-0.7.4/src/percept.c <==
/*
 * Copyright (C) 2002-2007 The Warp Rogue Team
 * Part of the Warp Rogue Project
 *
 * This software is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License.
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY.
 *
 * See the license.txt file for more details.
 */

/*
 * Module Name: Perception
 * Description: Character perception related things
 */

#define Uses_Character
#define Uses_Object
#define Uses_Util
#define Uses_Area
#define Uses_Faction
#define Uses_Dice
#define Uses_DynamicMessage
#define Uses_Terrain
#define Uses_Party
#define Uses_Death
#include "mheader.h"

#include "percept.h"


static bool	silent_character_noticed(const CHARACTER *, const CHARACTER *);

static bool	los_blocked(const AREA_POINT *);



/*
 * updates the passed character's perception data
 */
void update_perception_data(CHARACTER *character)
{
	CHARACTER *target;
	LIST_NODE *node;
	LIST *noticed_enemies;

	noticed_enemies = character->noticed_enemies;

	list_clear(noticed_enemies);

	for (node = area_character_list()->head;
		node != NULL;
		node = node->next) {

		target = (CHARACTER *)node->data;

		if (!hostility_between(character, target)) {

			continue;
		}

		if (!los(&character->location, &target->location)) {

			continue;
		}
		
		if (character_unnoticed(target)) {

			notice_check(target, true);
		}

		if (!character_unnoticed(target)) {

			list_add(noticed_enemies, target);

			continue;
		}
	}
}



/*
 * gives the unnoticed status to a character
 */
void activate_stealth(CHARACTER *character)
{

	character_set_flag(character, CF_UNNOTICED);
}



/*
 * removes the unnoticed status from a character
 */
void cancel_stealth(CHARACTER *character, const CHARACTER *noticed_by)
{

	if (noticed_by != NULL) {
		
		dynamic_message(MSG_NOTICE, noticed_by, character, 
			MOT_CHARACTER
		);
	}

	character_remove_flag(character, CF_UNNOTICED);
}



/*
 * returns true if the passed character has the unnoticed status
 */
bool character_unnoticed(const CHARACTER *character)
{

	return character_has_flag(character, CF_UNNOTICED);
}




/*
 * checks whether or not of a character has been noticed, 
 * and cancels the unnoticed status of the character if he
 * has been noticed
 */
void notice_check(CHARACTER *character, bool character_silent)
{
	LIST_NODE *node;
	
	if (character_killed(character)) {

		return;
	}

	for (node = area_character_list()->head;
		node != NULL;
		node = node->next) {
		CHARACTER *tested_character;

		tested_character = (CHARACTER *)node->data;

		if (!hostility_between(tested_character, character)) {

			continue;
		}

		if (!los(&tested_character->location, &character->location)) {

			continue;
		}
		
		if (!character_silent) {
		
			cancel_stealth(character, tested_character);
			
			break;
		}
			
		if (silent_character_noticed(character, tested_character)) {
			
			cancel_stealth(character, tested_character);
			
			break;
		}
	}
}



/*
 * returns true if the player's party can see the passed character
 */
bool party_can_see(const CHARACTER *character)
{
	LIST_NODE *node;
	
	if (character->party == PARTY_PLAYER) {
	
		return true;
	}
	
	if (character_unnoticed(character)) {
	
		return false;
	}

	for (node = party_player()->head;
		node != NULL;
		node = node->next) {
		const CHARACTER *party_member;

		party_member = (const CHARACTER *)node->data;

		if (los(&party_member->location, &character->location)) {

			return true;
		}
	}

	return false;
}



/*
 * returns true if the character is being watched by enemies
 */
bool enemies_are_watching(const CHARACTER *character)
{
	LIST_NODE *node;

	for (node = area_character_list()->head;
		node != NULL;
		node = node->next) {
		const CHARACTER *tested_character;

		tested_character = (const CHARACTER *)node->data;

		if (!hostility_between(tested_character, character)) {
			
			continue;
		}

		if (los(&tested_character->location, 
			&character->location)) {

			return true;
		}
	}

	return false;
}



/*
 * returns the nearest enemy the passed character is aware of
 */
CHARACTER * nearest_noticed_enemy(const CHARACTER *character)
{
	CHARACTER *nearest_enemy;
	AREA_DISTANCE nearest_enemy_distance;
	LIST_NODE *node;

	nearest_enemy_distance = AREA_DISTANCE_MAX;
	nearest_enemy = NULL;

	for (node = character->noticed_enemies->head;
		node != NULL;
		node = node->next) {
		CHARACTER *tested_character;
		AREA_DISTANCE tested_character_distance;

		tested_character = (CHARACTER *)node->data;
			
		if (!los(&tested_character->location,
			&character->location)) {
		
			continue;
		}

		if (character_unnoticed(tested_character)) {

			continue;
		}

		tested_character_distance = area_distance(
			&tested_character->location, &character->location
		);

		if (tested_character_distance < nearest_enemy_distance) {

			nearest_enemy = tested_character;
			nearest_enemy_distance = tested_character_distance;
		}
	}

	return nearest_enemy;
}



/*
 * returns true if there is a line of sight between the two passed points
 */
bool line_of_sight(const AREA_POINT *source, const AREA_POINT *target)
{
	int t, ax, ay, sx, sy, dx, dy;
	AREA_POINT point;

 	dx = target->x - source->x;
	dy = target->y - source->y;

	ax = abs(dx)<<1;
	ay = abs(dy)<<1;

	sx = util_sgn(dx);
	sy = util_sgn(dy);

	point = *source;

	if(ax > ay) {

		t = ay - (ax >> 1);

		do {

			if(t >= 0) {
				
           			point.y += sy;
				t -= ax;
			}

			point.x += sx;
			t += ay;

			if (area_points_equal(&point, target)) {

				return true;
			}

		} while (!los_blocked(&point));

		return false;

	} else {

		t = ax - (ay >> 1);

		do {

			if(t >= 0) {

				point.x += sx;
				t -= ay;
			}

			point.y += sy;
			t += ax;

			if(area_points_equal(&point, target)) {
				
				return true;
			}

		} while(!los_blocked(&point));

		return false;
	}
}



/*
 * checks whether or not of a silent character has been noticed
 */
static bool silent_character_noticed(const CHARACTER *c1, const CHARACTER *c2)
{
	NOT_USED(c1);

	if (!c2->perk[PK_ALERTNESS]) {

		return false;
	}

	if (d100_test_passed(c2->stat[S_IN].current)) {

		return true;
	}

	return false;
}



/*
 * returns true if there is something that blocks LOS at the passed location
 */
static bool los_blocked(const AREA_POINT *location)
{
	SECTOR *sector;

	sector = sector_at(location);

	if (sector->object != NULL) {

		if (object_has_attribute(sector->object,
			OA_BLOCKS_LINE_OF_SIGHT)) {

			return true;
		}
	}

	if (sector->terrain != NULL) {

		if (terrain_has_attribute(sector->terrain,
			TA_BLOCKS_LINE_OF_SIGHT)) {

			return true;
		}
	}

	return false;
}

==> wrogue-0.7.5.tar.bz2: wrogue-0.7.5/src/percept.c <==
/*
 * Copyright (C) 2002-2007 The Warp Rogue Team
 * Part of the Warp Rogue Project
 *
 * This software is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License.
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY.
 *
 * See the license.txt file for more details.
 */

/*
 * Module Name: Perception
 * Description: Character perception related things
 */

#define Uses_Character
#define Uses_Stats
#define Uses_Object
#define Uses_Util
#define Uses_Area
#define Uses_Faction
#define Uses_Dice
#define Uses_DynamicMessage
#define Uses_Terrain
#define Uses_Party
#define Uses_Death
#include "mheader.h"

#include "percept.h"


static bool	silent_character_noticed(const CHARACTER *, const CHARACTER *);

static bool	los_blocked(const AREA_POINT *);



/*
 * updates the passed character's perception data
 */
void update_perception_data(CHARACTER *character)
{
	CHARACTER *target;
	LIST_NODE *node;
	LIST *noticed_enemies;

	noticed_enemies = character->noticed_enemies;

	list_clear(noticed_enemies);

	for (node = area_character_list()->head;
		node != NULL;
		node = node->next) {

		target = (CHARACTER *)node->data;

		if (!hostility_between(character, target)) {

			continue;
		}

		if (!los(&character->location, &target->location)) {

			continue;
		}
		
		if (character_unnoticed(target)) {

			notice_check(target, true);
		}

		if (!character_unnoticed(target)) {

			list_add(noticed_enemies, target);

			continue;
		}
	}
}



/*
 * gives the unnoticed status to a character
 */
void activate_stealth(CHARACTER *character)
{

	character_set_flag(character, CF_UNNOTICED);
}



/*
 * removes the unnoticed status from a character
 */
void cancel_stealth(CHARACTER *character, const CHARACTER *noticed_by)
{

	if (noticed_by != NULL) {
		
		dynamic_message(MSG_NOTICE, noticed_by, character, 
			MOT_CHARACTER
		);
	}

	character_remove_flag(character, CF_UNNOTICED);
}



/*
 * returns true if the passed character has the unnoticed status
 */
bool character_unnoticed(const CHARACTER *character)
{

	return character_has_flag(character, CF_UNNOTICED);
}




/*
 * checks whether or not of a character has been noticed, 
 * and cancels the unnoticed status of the character if he
 * has been noticed
 */
void notice_check(CHARACTER *character, bool character_silent)
{
	LIST_NODE *node;
	
	if (character_killed(character)) {

		return;
	}

	for (node = area_character_list()->head;
		node != NULL;
		node = node->next) {
		CHARACTER *tested_character;

		tested_character = (CHARACTER *)node->data;

		if (!hostility_between(tested_character, character)) {

			continue;
		}

		if (!los(&tested_character->location, &character->location)) {

			continue;
		}
		
		if (!character_silent) {
		
			cancel_stealth(character, tested_character);
			
			break;
		}
			
		if (silent_character_noticed(character, tested_character)) {
			
			cancel_stealth(character, tested_character);
			
			break;
		}
	}
}



/*
 * returns true if the player's party can see the passed character
 */
bool party_can_see(const CHARACTER *character)
{
	LIST_NODE *node;
	
	if (character->party == PARTY_PLAYER) {
	
		return true;
	}
	
	if (character_unnoticed(character)) {
	
		return false;
	}

	for (node = party_player()->head;
		node != NULL;
		node = node->next) {
		const CHARACTER *party_member;

		party_member = (const CHARACTER *)node->data;

		if (los(&party_member->location, &character->location)) {

			return true;
		}
	}

	return false;
}



/*
 * returns true if the character is being watched by enemies
 */
bool enemies_are_watching(const CHARACTER *character)
{
	LIST_NODE *node;

	for (node = area_character_list()->head;
		node != NULL;
		node = node->next) {
		const CHARACTER *tested_character;

		tested_character = (const CHARACTER *)node->data;

		if (!hostility_between(tested_character, character)) {
			
			continue;
		}

		if (los(&tested_character->location, 
			&character->location)) {

			return true;
		}
	}

	return false;
}



/*
 * returns the nearest enemy the passed character is aware of
 */
CHARACTER * nearest_noticed_enemy(const CHARACTER *character)
{
	CHARACTER *nearest_enemy;
	AREA_DISTANCE nearest_enemy_distance;
	LIST_NODE *node;

	nearest_enemy_distance = AREA_DISTANCE_MAX;
	nearest_enemy = NULL;

	for (node = character->noticed_enemies->head;
		node != NULL;
		node = node->next) {
		CHARACTER *tested_character;
		AREA_DISTANCE tested_character_distance;

		tested_character = (CHARACTER *)node->data;
			
		if (!los(&tested_character->location,
			&character->location)) {
		
			continue;
		}

		if (character_unnoticed(tested_character)) {

			continue;
		}

		tested_character_distance = area_distance(
			&tested_character->location, &character->location
		);

		if (tested_character_distance < nearest_enemy_distance) {

			nearest_enemy = tested_character;
			nearest_enemy_distance = tested_character_distance;
		}
	}

	return nearest_enemy;
}



/*
 * returns true if there is an unnoticed enemy at the passed location
 */
bool unnoticed_enemy_at(const CHARACTER *character, const AREA_POINT *location)
{
	const CHARACTER *character_at_location;

	character_at_location = character_at(location);

	if (character_at_location == NULL) {
	
		return false;
	}
	
	if (character_unnoticed(character_at_location) &&
		hostility_between(character, character_at_location)) {

		return true;
	}

	return false;
}



/*
 * returns true if there is a line of sight between the two passed points
 */
bool line_of_sight(const AREA_POINT *source, const AREA_POINT *target)
{
	int t, ax, ay, sx, sy, dx, dy;
	AREA_POINT point;

 	dx = target->x - source->x;
	dy = target->y - source->y;

	ax = abs(dx)<<1;
	ay = abs(dy)<<1;

	sx = sgn(dx);
	sy = sgn(dy);

	point = *source;

	if(ax > ay) {

		t = ay - (ax >> 1);

		do {

			if(t >= 0) {
				
           			point.y += sy;
				t -= ax;
			}

			point.x += sx;
			t += ay;

			if (area_points_equal(&point, target)) {

				return true;
			}

		} while (!los_blocked(&point));

		return false;

	} else {

		t = ax - (ay >> 1);

		do {

			if(t >= 0) {

				point.x += sx;
				t -= ay;
			}

			point.y += sy;
			t += ax;

			if(area_points_equal(&point, target)) {
				
				return true;
			}

		} while(!los_blocked(&point));

		return false;
	}
}



/*
 * checks whether or not of a silent character has been noticed
 */
static bool silent_character_noticed(const CHARACTER *c1, const CHARACTER *c2)
{
	NOT_USED(c1);

	if (c2->perk[PK_ALERTNESS] || 
		d100_test_passed(c2->stat[S_IN].current)) {

		return true;
	}

	return false;
}



/*
 * returns true if there is something that blocks LOS at the passed location
 */
static bool los_blocked(const AREA_POINT *location)
{
	SECTOR *sector;

	sector = sector_at(location);

	if (sector->object != NULL) {

		if (object_has_attribute(sector->object,
			OA_BLOCKS_LINE_OF_SIGHT)) {

			return true;
		}
	}

	if (sector->terrain != NULL) {

		if (terrain_has_attribute(sector->terrain,
			TA_BLOCKS_LINE_OF_SIGHT)) {

			return true;
		}
	}

	return false;
}
